# -*- coding:utf-8 -*-

import argparse

from os.path import join, isdir, basename
from os import listdir

import numpy as np

from .eer_stats import calculate_roc, calculate_roc_hist, calculate_roc_auc,\
    get_fmr_op, get_fnmr_op, get_eer_values, Stats, get_decidability_value,\
    get_youden_index, get_matthews_ccoef
from .reports import generate_eer_report, plot_eer_stats

__copyright__ = 'Copyright 2017'
__author__ = u'Bsc. Manuel Aguado Martínez'


def __get_files(base_path, arg_val):
    """ Get the file names from the argument value

    @param base_path: The path where the files are suppose to be
    @type base_path: str
    @param arg_val: The argument value of genuine or impostor scores files
    @type arg_val: str
    """
    if isdir(join(base_path, arg_val)):
        files = sorted(listdir(join(base_path, arg_val)))
        return [join(arg_val, f) for f in files]
    else:
        return [f.strip() for f in arg_val.split(',')]


def __get_score(line):
    """Get the score value from an input score file line

    @param line: An input score file line
    @type line: str

    @returns: The parsed score
    @rtype: float
    """
    sline = line.strip().split(' ')
    return float(sline[-1])


def get_eer_info_cmd():
    ap = argparse.ArgumentParser()
    ap.add_argument("-p", "--path", required=False, default='.',
                    help="Path to scores files. (Default='.')")
    ap.add_argument("-i", "--iscores_files", required=True,
                    help="Impostor scores file names. If there is more than "
                         " one file, the file names must be separated by comma"
                         ". Additionally, instead of the file names a"
                         " directory relative to PATH could be specified."
                         " In the last case, it is hardly encouraged that"
                         " genuine score files are specified in the same way"
                         " and corresponding pairs of scores (genuine, "
                         " impostor) have the same name")
    ap.add_argument("-g", "--gscores_files", required=True,
                    help="Genuine scores file names. If there is more than "
                         " one file, the file names must be separated by comma"
                         ". Additionally, instead of the file names a"
                         " directory relative to PATH could be specified."
                         " In the last case, it is hardly encouraged that"
                         " impostor score files are specified in the same way"
                         " and corresponding pairs of scores (genuine, "
                         " impostor) have the same name")
    ap.add_argument("-e", "--experiment_ids", required=False,
                    help="Experiment IDs separated by comma. If not"
                         " specified genuine score file names will be used to"
                         " identified each experiment")
    ap.add_argument("-ht", "--hist", required=False, action='store_true',
                    help="Indicates that the impostor file is in"
                         " histogram format")
    ap.add_argument("-ds", "--ds_scores", required=False, action='store_true',
                    help="Indicates whether the input scores are dissimilarity"
                         " scores")
    ap.add_argument("-hb", "--distribution_bins", required=False, default=100,
                    help="The number of bins to compute scores distribution."
                         " Will be ignored if -ht is passed as parameter"
                         " (default=100)")
    ap.add_argument("-np", "--no_plots", required=False, action='store_true',
                    help="Indicates whether to not plot the results")
    ap.add_argument("-lg", "--log_plt", required=False, action='store_true',
                    help="Indicates whether to plot the DET and ROC curves "
                         "in logarithmic scale")
    ap.add_argument("-s", "--save_plots", required=False, action='store_true',
                    help="Indicates whether to save the plots instead of"
                         " showing them")
    ap.add_argument("-sp", "--save_path", required=False, default='',
                    help="Path to save the plots (if -s was specified)"
                         " and stats report")
    ap.add_argument("-sf", "--save_format", required=False, default='png',
                    help="Format to save the plots in the cases where the"
                         " option -s was specified. Valid formats are: "
                         "(png, pdf, ps, eps and svg)")
    ap.add_argument("-sr", "--save_dpi", required=False, default=None,
                    help="Plots resolution (dots per inch) in the cases"
                         " where the option -s was specified. If not given"
                         " it will default to the value savefig.dpi in the"
                         " matplotlibrc file")
    ap.add_argument("-lw", "--line_width", required=False, default=3,
                    help="The width of the plotted curves (default=3)")
    ap.add_argument("-ls", "--legend_font_size", required=False, default=15,
                    help="The size of the legend font (default=15)")
    args = ap.parse_args()

    # Parsing arguments
    gscores_files = __get_files(args.path, args.gscores_files)
    iscores_files = __get_files(args.path, args.iscores_files)

    experiment_ids = ([e.strip() for e in args.experiment_ids.split(',')]
                      if args.experiment_ids else
                      [basename(name) for name in gscores_files])

    experiments = zip(gscores_files, iscores_files, experiment_ids)

    # Plot arguments
    line_width = int(args.line_width)
    lgf_size = int(args.legend_font_size)
    dpi = None if args.save_dpi is None else int(args.save_dpi)
    ext = '.' + args.save_format
    bins = int(args.distribution_bins)

    # Experiment stats
    stats = []
    ids = []

    for exp in experiments:
        # Loading scores
        print('%s: Loading genuine scores file...' % exp[2])
        with open(join(args.path, exp[0])) as tf:
            gen_scores = [__get_score(line) for line in tf]

        print('%s: Loading impostor scores file...' % exp[2])
        with open(join(args.path, exp[1])) as tf:
            imp_scores = [__get_score(line) for line in tf]

        print('%s: Calculating stats...' % exp[2])
        exp_stats = get_eer_stats(gen_scores, imp_scores,
                                  args.hist, args.ds_scores)
        stats.append(exp_stats)
        ids.append(exp[2])

    # Generating reports
    print('Generating report...')

    generate_eer_report(stats, ids, join(args.save_path, 'pyeer_report.csv'))

    if not args.no_plots:
        print('Plotting...')
        plot_eer_stats(stats, ids, line_width, args.hist, bins, lgf_size,
                       args.log_plt, args.save_plots, dpi, args.save_path,
                       ext)


def get_eer_stats(gen_scores, imp_scores, hformat=False, ds_scores=False):
    """Calculates EER associated statistics

    Keyword Arguments:
    @param gen_scores: The genuine scores
    @type gen_scores: list
    @param imp_scores: The impostor scores
    @type imp_scores: list
    @param id: An id for the experiment
    @type id: str
    @param hformat: Indicates whether the impostor scores are in histogram
        format
    @type hformat: bool
    @param ds_scores: Indicates whether the input scores are dissimilarity
        scores
    @type ds_scores: bool
    """

    if hformat:
        # Calculating probabilities histogram format
        roc_info = calculate_roc_hist(gen_scores, imp_scores,
                                      ds_scores, rates=False)
        gnumber = float(len(gen_scores))
        inumber = float(sum(imp_scores))
    else:
        # Calculating probabilities using scores as thrs
        roc_info = calculate_roc(gen_scores, imp_scores,
                                 ds_scores, rates=False)
        gnumber = len(gen_scores)
        inumber = len(imp_scores)

    # Unboxing probability rates and info
    thrs, fm, fnm = roc_info
    fmr = fm / inumber
    fnmr = fnm / gnumber

    # Estimating EER
    eer_low, eer_high, eer = get_eer_values(fmr, fnmr)

    # Estimating FMR operating points
    fmr0 = get_fmr_op(fmr, fnmr, 0)
    fmr1000 = get_fmr_op(fmr, fnmr, 0.001)
    fmr100 = get_fmr_op(fmr, fnmr, 0.01)
    fmr20 = get_fmr_op(fmr, fnmr, 0.05)
    fmr10 = get_fmr_op(fmr, fnmr, 0.1)

    # Estimating FNMR operating points
    fnmr0 = get_fnmr_op(fmr, fnmr, 0)

    # Calculating distributions mean and variance
    gmean = np.mean(gen_scores)
    gstd = np.std(gen_scores)

    if hformat:
        nscores = sum(imp_scores)
        nscores_prob = np.array(imp_scores) / nscores
        scores = np.arange(len(imp_scores))

        imean = (scores * nscores_prob).sum()
        istd = np.sqrt(((scores - imean) ** 2 * nscores_prob).sum())
    else:
        imean = np.mean(imp_scores)
        istd = np.std(imp_scores)

    dec = get_decidability_value(gmean, gstd, imean, istd)

    # Calculating area under the ROC curve
    auc = calculate_roc_auc(fmr, fnmr)

    j_index, j_index_th = get_youden_index(fmr, fnmr)
    j_index_th = thrs[j_index_th]

    mccoef, mccoef_th = get_matthews_ccoef(fm, fnm, gnumber, inumber)
    mccoef_th = thrs[mccoef_th]

    # Stacking stats
    return Stats(thrs=thrs, fmr=fmr, fnmr=fnmr, auc=auc, eer=eer,
                 fmr0=fmr0, fmr100=fmr100, fmr1000=fmr1000,
                 fmr20=fmr20, fmr10=fmr10, fnmr0=fnmr0,
                 gen_scores=gen_scores, imp_scores=imp_scores,
                 gmean=gmean, gstd=gstd, imean=imean, istd=istd,
                 eer_low=eer_low, eer_high=eer_high, decidability=dec,
                 j_index=j_index, j_index_th=j_index_th,
                 mccoef=mccoef, mccoef_th=mccoef_th)



# -*- coding:utf-8 -*-

import operator

from collections import namedtuple

import numpy as np


__copyright__ = 'Copyright 2017'
__author__ = u'Bsc. Manuel Aguado Martínez'


Stats = namedtuple('Stats', [

    # 'exp_id',  # Exp id

    # Rate curves
    'thrs',  # Thresholds
    'fmr',  # False match rates
    'fnmr',  # False non-match rates
    'auc',  # Area under the ROC curve
    'j_index',  # Youden's index
    'j_index_th',  # Youden's index threshold
    'mccoef',  # Matthew correlation coefficient
    'mccoef_th',  # Matthew correlation coefficient threshold

    # Operation points
    'fmr0',  # Zero false math rate
    'fmr1000',  # 1000 false match rate
    'fmr100',  # 100 false match rate
    'fmr20',  # 20 false match rate
    'fmr10',  # 10 false match rate
    'fnmr0',  # 0 false non-match rate

    # Scores distributions
    'gen_scores',  # Genuine scores
    'imp_scores',  # Impostor scores
    'gmean',  # Genuine scores mean
    'gstd',  # Genuine scores standard deviation
    'imean',  # Impostor scores mean
    'istd',  # Impostor scores standard deviation
    'decidability',  # Decidability score

    # Values of EER
    'eer',  # Equal error rate
    'eer_low',  # Equal error rate (low)
    'eer_high'  # Equal error rate (high)
])


def calculate_roc_hist(gscores, iscores, ds_scores=False, rates=True):
    """Calculates FMR, FNMR for impostor scores in histogram format

    @param gscores: Genuine matching scores
    @type gscores: list
    @param iscores: Impostor matching scores
    @type giscores: list
    @param ds_scores: Indicates whether input scores are
        dissimilarity scores
    @type ds_scores: bool
    @param rates: Indicates whether to return error rates instead
        of error values
    @type rates: bool

    @return: (thresholds, FMR, FNMR) or (thresholds, FM, FNM)
    @rtype: tuple
    """
    match_thr = 0
    maximum_thr = max([max(gscores), len(gscores),
                       max(iscores), len(iscores)])

    gscores_number = float(len(gscores))
    iscores_number = float(sum(iscores))
    iscores = np.cumsum(iscores)

    thresholds = []
    fm_rates = []
    fnm_rates = []

    while match_thr <= maximum_thr:
        if 0 < match_thr < len(iscores):
            if ds_scores:
                false_match = iscores[match_thr - 1]
            else:
                false_match = iscores_number - iscores[match_thr - 1]
        elif match_thr == 0:
            if ds_scores:
                false_match = 0
            else:
                false_match = iscores_number  # Accepting everyone
        else:
            if ds_scores:
                false_match = iscores_number
            else:
                false_match = 0

        fm_rates.append(false_match)

        # List convertion for Python3 compatibility
        if ds_scores:
            fnm = len(list(filter(lambda s: s > match_thr, gscores)))
        else:
            fnm = len(list(filter(lambda s: s < match_thr, gscores)))
        fnm_rates.append(fnm)

        thresholds.append(match_thr)
        match_thr += 1

    if rates:
        fm_rates = np.array(fm_rates) / iscores_number
        fnm_rates = np.array(fnm_rates) / gscores_number
    else:
        fm_rates = np.array(fm_rates)
        fnm_rates = np.array(fnm_rates)

    return thresholds, fm_rates, fnm_rates


def calculate_roc(gscores, iscores, ds_scores=False, rates=True):
    """Calculates FMR, FNMR

    @param gscores: Genuine matching scores
    @type gscores: list
    @param iscores: Impostor matching scores
    @type giscores: list
    @param ds_scores: Indicates whether input scores are
        dissimilarity scores
    @type ds_scores: bool
    @param rates: Indicates whether to return error rates instead
        of error values
    @type rates: bool

    @return: (thresholds, FMR, FNMR) or (thresholds, FM, FNM)
    @rtype: tuple
    """
    if ds_scores:
        gscores = np.array(gscores) * -1
        iscores = np.array(iscores) * -1

    gscores_number = len(gscores)
    iscores_number = len(iscores)

    # Labeling genuine scores as 1 and impostor scores as 0
    gscores = zip(gscores, [1] * gscores_number)
    iscores = zip(iscores, [0] * iscores_number)

    # Python3 compatibility
    gscores = list(gscores)
    iscores = list(iscores)

    # Stacking scores
    scores = np.array(sorted(gscores + iscores, key=operator.itemgetter(0)))
    cumul = np.cumsum(scores[:, 1])

    # Grouping scores
    thresholds, u_indices = np.unique(scores[:, 0], return_index=True)

    # Calculating FNM and FM distributions
    fnm = cumul[u_indices] - scores[u_indices][:, 1]  # rejecting s < t
    fm = iscores_number - (u_indices - fnm)

    # Calculating FMR and FNMR
    if rates:
        fnm_rates = fnm / gscores_number
        fm_rates = fm / iscores_number
    else:
        fnm_rates = fnm
        fm_rates = fm

    if ds_scores:
        return thresholds * -1, fm_rates, fnm_rates

    return thresholds, fm_rates, fnm_rates


def calculate_roc_auc(fmr, fnmr):
    """Calculates the area under a ROC curve

    @param fmr: False Match Rates
    @type fmr: ndarray
    @param fnmr: False Non-Match Rates
    @type fnmr: ndarray

    @returns: Area under the ROC curve
    @rtype: float
    """
    x1 = fmr[:-1]
    x2 = fmr[1:]

    tpr = 1 - fnmr
    y1 = tpr[:-1]
    y2 = tpr[1:]

    return ((x1 - x2) * (y1 + (y2 - y1) / 2)).sum()


def get_fnmr_op(fmr, fnmr, op):
    """Returns the value of the given FNMR operating point

    Definition:
    ZeroFNMR: is defined as the lowest FMR at which no non-false matches occur.

    Others FNMR operating points are defined in a similar way.

    @param fmr: False Match Rates
    @type fmr: ndarray
    @param fnmr: False Non-Match Rates
    @type fnmr: ndarray
    @param op: Operating point
    @type op: float

    @returns: The lowest FMR at which the probability of FNMR == op
    @rtype: float
    """
    temp = abs(fnmr - op)
    min_val = np.min(temp)
    index = np.where(temp == min_val)[0][-1]
    return fmr[index]


def get_fmr_op(fmr, fnmr, op):
    """Returns the value of the given FMR operating point

    Definition:
    ZeroFMR: is defined as the lowest FNMR at which no false matches occur.

    Others FMR operating points are defined in a similar way.

    @param fmr: False Match Rates
    @type fmr: ndarray
    @param fnmr: False Non-Match Rates
    @type fnmr: ndarray
    @param op: Operating point
    @type op: float

    @returns: The lowest FNMR at which the probability of FMR == op
    @rtype: float
    """
    index = np.argmin(abs(fmr - op))
    return fnmr[index]


def get_eer_values(fmr, fnmr):
    """Returns the value of the Equal Error Rate

    Equal Error Rate (EER): is the point where FNMR(t) = FMR(t).
    In practice the score distribution are not continuous so and interval
    is returned instead. The EER value will be set as the midpoint of
    this interval.

    The interval will be defined as:
    [EERlow, EERhigh] = min(fnmr[t], fmr[t]), max(fnmr[t], fmr[t])
    where t = argmin(abs(fnmr - fmr))

    The EER value is computed as (EERlow + EERhigh) / 2

    Reference:
    Maio, D., Maltoni, D., Cappelli, R., Wayman, J. L., & Jain, A. K. (2002).
    FVC2000: Fingerprint verification competition. IEEE Transactions on
    Pattern Analysis and Machine Intelligence, 24(3), 402-412.

    @param fmr: False Match Rates (FMR)
    @type fmr: ndarray
    @param fnmr: False Non-Match Rates (FNMR)
    @type fnmr: ndarray
    @param op: Operation point
    @type op: float

    @returns: EERlow, EERhigh, EER
    @rtype: tuple
    """
    diff = fmr - fnmr
    t2 = np.where(diff <= 0)[0][0]
    t1 = t2 - 1 if diff[t2] != 0 and t2 != 0 else t2

    if fmr[t1] + fnmr[t1] <= fmr[t2] + fnmr[t2]:
        return fnmr[t1], fmr[t1], (fnmr[t1] + fmr[t1]) / 2
    else:
        return fmr[t2], fnmr[t2], (fnmr[t2] + fmr[t2]) / 2

    # index = np.argmin(abs(fmr - fnmr))
    # eer = np.abs(fnmr[index] + fmr[index]) / 2.0
    # return min(fnmr[index], fmr[index]), max(fnmr[index], fmr[index]), eer


def get_decidability_value(gmean, gstd, imean, istd):
    """ The decidability score (d') or decision-making power used in NICE:II
    evauation protocol for iris, it is a measure of the separation between
    genuine and impostor distributions. Higher d' values indicate better
    separation. It is also called the sensitivity index.

    @param gmean: The mean value of the genuine scores
    @type gmean: float
    @param imean: The mean value of the impostor scores
    @type imean: float
    @param gstd: The standard deviation value of the genuine scores
    @type gstd: float
    @param istd: The standard deviation value of the impostor scores
    @type istd: float

    @returns: The decidability value
    @rtype: float
    """
    return abs(gmean - imean) / np.sqrt(0.5 * (gstd ** 2 + istd ** 2))


def get_youden_index(fmr, fnmr):
    """Computes the Youden's index and the corresponding threshold

    Reference:
    Youden, W. J. (1950). "Index of rating diagnostic tests". Cancer. 3: 32-35

    @param fmr: False Match Rates (FMR)
    @type fmr: ndarray
    @param fnmr: False Non-Match Rates (FNMR)
    @type fnmr: ndarray

    @returns: Youden's Index, threshold
    @rtype: tuple
    """
    # J = sensitivity + specificity -1
    # sensitivity = 1 - fnmr
    # specificity = 1 - fmr
    # J = 1 - fnmr + 1 - fmr - 1
    # J = 1 - fnmr - fmr
    j = 1 - fnmr - fmr
    th = np.argmax(j)
    return j[th], th


def get_matthews_ccoef(fm, fnm, gnumber, inumber):
    """Estimate the maximum Matthews Correlation Coefficient
    given the values of false match (false positives) and false
    non-match (false negatives) for all possible thresholds

    @param fm: False Positives for all possible thresholds
    @type fm: ndarray
    @param fnm: False Negatives for all possible thresholds
    @type fnm: ndarray
    @param gnumber: The number of positive samples
    @type gnumber: int
    @param inumber: The number of negative samples
    @type inumber: int

    @returns: (Matthews Correlation Coefficient, threshold)
    @rtype: tuple
    """
    tn = inumber - fm
    tp = gnumber - fnm

    numerator = tp * tn - fm * fnm
    denominator = np.sqrt((tp + fm) * (tp + fnm) * (tn + fm) * (tn + fnm))

    denominator[denominator == 0] = 1

    all_mcc = numerator / denominator

    th = np.argmax(all_mcc)
    return all_mcc[th], th



# -*- coding:utf-8 -*-

import operator

from collections import namedtuple

__copyright__ = 'Copyright 2017'
__author__ = u'Manuel Aguado Martínez'

TEMPLATE_POS = 0
SCORE_POS = 1


CMCstats = namedtuple('CMCstats', ['exp_id',  # Exp id
                                   'ranks',  # Rank values
                                   ])


def load_scores_from_file(scores_filename, true_pairs_filename,
                          ds_scores=False):
    """Loads the match information from the files.

    @param scores_filename: The scores file address. One score per
        line with the following format: (query template score)
    @type scores_filename: str
    @param true_pairs_filename: The true pairs file address. Each line
        indicates the corresponding template of each query. Must have
        the following format: (query true_template)
    @type true_pairs_filename: str
    @param ds_scores: Indicates whether te input scores are dissimilarity
        scores.
    @type ds_scores: bool

    @returns: A dictionary {key=query, value=QueryMatchInfo}
    @rtype: dict
    """
    matching_scores = {}

    with open(true_pairs_filename) as tpf:
        for line in tpf:
            query, template = line.split(' ', 1)
            matching_scores[query] = (template, [])

    with open(scores_filename) as sf:
        for line in sf:
            query, template, score = line.split(' ')[:3]
            matching_scores[query][SCORE_POS].append((template, float(score)))

    for query_match_info in matching_scores.values():
        query_match_info[SCORE_POS].sort(key=operator.itemgetter(SCORE_POS),
                                         reverse=not ds_scores)

    return matching_scores


def get_cmc_curve(scores, max_rank):
    """Calculate the values of a CMC curve

    @param scores: The dictionary returned by the function
        load_scores_from_file or a similar one.
    @type scores: dict
    @param max_rank: The maximum rank to calculate the penetration coefficient.
    @type max_rank : int

    @return: A list with the rank values.
    @rtype: list
    """
    ranks_values = [0.0] * (max_rank + 1)
    queries_total = len(scores)

    for r in range(max_rank):
        in_rank = 0.0
        for query_match_info in scores.values():
            match = query_match_info[SCORE_POS][r]
            true_template = query_match_info[TEMPLATE_POS].strip()
            if match[TEMPLATE_POS] == true_template:
                in_rank += 1
        ranks_values[r + 1] = in_rank / queries_total + ranks_values[r]

    return ranks_values[1:]
